@@ -21,73 +21,25 @@
 (* Extensive Rewrite: Hongbo Zhang: University of Pennsylvania *)
 (* TODO more fine-grained precedence pretty-printing *)
 
-open Ast_414
+(* This file was copied from OCaml 5.2's pprintast.ml and modified in the
+   following ways:
+   - Added [open Ast_502] before other global opens
+   - Replaced [Lexer.is_keyword] with [Keyword.is_keyword] for compat with
+     Ocaml < 5.2.
+   - Added [class_signature] and [type_declaration] entry points at the end.
+   - Added a custom case to `binding` to print specific instances of
+     [Ppat_constraint (p, typ)] in [value_binding] patterns as if they were encoded
+     using the new [pvb_constraint] field instead of producing incorrect syntax as
+     the compiler version does.
+*)
+
+open Ast_502
 open Asttypes
 open Format
 open Location
 open Longident
 open Parsetree
 
-module Option = struct
-  let value t ~default = match t with None -> default | Some x -> x
-end
-
-let varify_type_constructors var_names t =
-  let check_variable vl loc v =
-    if List.mem v vl then
-      Location.raise_errorf ~loc "variable in scope syntax error: %s" v
-  in
-  let var_names = List.map (fun v -> v.txt) var_names in
-  let rec loop t =
-    let desc =
-      match t.ptyp_desc with
-      | Ptyp_any -> Ptyp_any
-      | Ptyp_var x ->
-          check_variable var_names t.ptyp_loc x;
-          Ptyp_var x
-      | Ptyp_arrow (label, core_type, core_type') ->
-          Ptyp_arrow (label, loop core_type, loop core_type')
-      | Ptyp_tuple lst -> Ptyp_tuple (List.map loop lst)
-      | Ptyp_constr ({ txt = Longident.Lident s }, []) when List.mem s var_names
-        ->
-          Ptyp_var s
-      | Ptyp_constr (longident, lst) ->
-          Ptyp_constr (longident, List.map loop lst)
-      | Ptyp_object (lst, o) -> Ptyp_object (List.map loop_object_field lst, o)
-      | Ptyp_class (longident, lst) -> Ptyp_class (longident, List.map loop lst)
-      | Ptyp_alias (core_type, string) ->
-          check_variable var_names t.ptyp_loc string;
-          Ptyp_alias (loop core_type, string)
-      | Ptyp_variant (row_field_list, flag, lbl_lst_option) ->
-          Ptyp_variant
-            (List.map loop_row_field row_field_list, flag, lbl_lst_option)
-      | Ptyp_poly (string_lst, core_type) ->
-          List.iter
-            (fun v -> check_variable var_names t.ptyp_loc v.txt)
-            string_lst;
-          Ptyp_poly (string_lst, loop core_type)
-      | Ptyp_package (longident, lst) ->
-          Ptyp_package (longident, List.map (fun (n, typ) -> (n, loop typ)) lst)
-      | Ptyp_extension (s, arg) -> Ptyp_extension (s, arg)
-    in
-    { t with ptyp_desc = desc }
-  and loop_row_field field =
-    let prf_desc =
-      match field.prf_desc with
-      | Rtag (label, flag, lst) -> Rtag (label, flag, List.map loop lst)
-      | Rinherit t -> Rinherit (loop t)
-    in
-    { field with prf_desc }
-  and loop_object_field field =
-    let pof_desc =
-      match field.pof_desc with
-      | Otag (label, t) -> Otag (label, loop t)
-      | Oinherit t -> Oinherit (loop t)
-    in
-    { field with pof_desc }
-  in
-  loop t
-
 let prefix_symbols = [ '!'; '?'; '~' ]
 
 let infix_symbols =
@@ -147,24 +99,27 @@ let needs_parens txt =
 (* some infixes need spaces around parens to avoid clashes with comment
    syntax *)
 let needs_spaces txt = first_is '*' txt || last_is '*' txt
-let string_loc ppf x = fprintf ppf "%s" x.txt
 
-(* add parentheses to binders when they are in fact infix or prefix operators *)
-let protect_ident ppf txt =
+(* Turn an arbitrary variable name into a valid OCaml identifier by adding \#
+  in case it is a keyword, or parenthesis when it is an infix or prefix
+  operator. *)
+let ident_of_name ppf txt =
   let format : (_, _, _) format =
-    if not (needs_parens txt) then "%s"
+    if Keyword.is_keyword txt then "\\#%s"
+    else if not (needs_parens txt) then "%s"
     else if needs_spaces txt then "(@;%s@;)"
     else "(%s)"
   in
   fprintf ppf format txt
 
+let ident_of_name_loc ppf s = ident_of_name ppf s.txt
+
 let protect_longident ppf print_longident longprefix txt =
-  let format : (_, _, _) format =
-    if not (needs_parens txt) then "%a.%s"
-    else if needs_spaces txt then "%a.(@;%s@;)"
-    else "%a.(%s)"
-  in
-  fprintf ppf format print_longident longprefix txt
+  if not (needs_parens txt) then
+    fprintf ppf "%a.%a" print_longident longprefix ident_of_name txt
+  else if needs_spaces txt then
+    fprintf ppf "%a.(@;%s@;)" print_longident longprefix txt
+  else fprintf ppf "%a.(%s)" print_longident longprefix txt
 
 type space_formatter = (unit, Format.formatter, unit) format
 
@@ -184,11 +139,15 @@ type construct =
   | `nil
   | `normal
   | `simple of Longident.t
-  | `tuple ]
+  | `tuple
+  | `btrue
+  | `bfalse ]
 
 let view_expr x =
   match x.pexp_desc with
   | Pexp_construct ({ txt = Lident "()"; _ }, _) -> `tuple
+  | Pexp_construct ({ txt = Lident "true"; _ }, _) -> `btrue
+  | Pexp_construct ({ txt = Lident "false"; _ }, _) -> `bfalse
   | Pexp_construct ({ txt = Lident "[]"; _ }, _) -> `nil
   | Pexp_construct ({ txt = Lident "::"; _ }, Some _) ->
       let rec loop exp acc =
@@ -215,17 +174,20 @@ let view_expr x =
   | _ -> `normal
 
 let is_simple_construct : construct -> bool = function
-  | `nil | `tuple | `list _ | `simple _ -> true
+  | `nil | `tuple | `list _ | `simple _ | `btrue | `bfalse -> true
   | `cons _ | `normal -> false
 
 let pp = fprintf
 
-type ctxt = { pipe : bool; semi : bool; ifthenelse : bool }
+type ctxt = { pipe : bool; semi : bool; ifthenelse : bool; functionrhs : bool }
+
+let reset_ctxt =
+  { pipe = false; semi = false; ifthenelse = false; functionrhs = false }
 
-let reset_ctxt = { pipe = false; semi = false; ifthenelse = false }
 let under_pipe ctxt = { ctxt with pipe = true }
 let under_semi ctxt = { ctxt with semi = true }
 let under_ifthenelse ctxt = { ctxt with ifthenelse = true }
+let under_functionrhs ctxt = { ctxt with functionrhs = true }
 (*
 let reset_semi ctxt = { ctxt with semi=false }
 let reset_ifthenelse ctxt = { ctxt with ifthenelse=false }
@@ -233,14 +195,14 @@ let reset_pipe ctxt = { ctxt with pipe=false }
 *)
 
 let list :
-      'a.
-      ?sep:space_formatter ->
-      ?first:space_formatter ->
-      ?last:space_formatter ->
-      (Format.formatter -> 'a -> unit) ->
-      Format.formatter ->
-      'a list ->
-      unit =
+    'a.
+    ?sep:space_formatter ->
+    ?first:space_formatter ->
+    ?last:space_formatter ->
+    (Format.formatter -> 'a -> unit) ->
+    Format.formatter ->
+    'a list ->
+    unit =
  fun ?sep ?first ?last fu f xs ->
   let first = match first with Some x -> x | None -> ("" : _ format6)
   and last = match last with Some x -> x | None -> ("" : _ format6)
@@ -264,13 +226,13 @@ let list :
   aux f xs
 
 let option :
-      'a.
-      ?first:space_formatter ->
-      ?last:space_formatter ->
-      (Format.formatter -> 'a -> unit) ->
-      Format.formatter ->
-      'a option ->
-      unit =
+    'a.
+    ?first:space_formatter ->
+    ?last:space_formatter ->
+    (Format.formatter -> 'a -> unit) ->
+    Format.formatter ->
+    'a option ->
+    unit =
  fun ?first ?last fu f a ->
   let first = match first with Some x -> x | None -> ("" : _ format6)
   and last = match last with Some x -> x | None -> ("" : _ format6) in
@@ -282,14 +244,14 @@ let option :
       pp f last
 
 let paren :
-      'a.
-      ?first:space_formatter ->
-      ?last:space_formatter ->
-      bool ->
-      (Format.formatter -> 'a -> unit) ->
-      Format.formatter ->
-      'a ->
-      unit =
+    'a.
+    ?first:space_formatter ->
+    ?last:space_formatter ->
+    bool ->
+    (Format.formatter -> 'a -> unit) ->
+    Format.formatter ->
+    'a ->
+    unit =
  fun ?(first = ("" : _ format6)) ?(last = ("" : _ format6)) b fu f x ->
   if b then (
     pp f "(";
@@ -300,7 +262,7 @@ let paren :
   else fu f x
 
 let rec longident f = function
-  | Lident s -> protect_ident f s
+  | Lident s -> ident_of_name f s
   | Ldot (y, s) -> protect_longident f longident y s
   | Lapply (y, s) -> pp f "%a(%a)" longident y longident s
 
@@ -336,15 +298,18 @@ let private_flag f = function Public -> () | Private -> pp f "private@ "
 let iter_loc f ctxt { txt; loc = _ } = f ctxt txt
 let constant_string f s = pp f "%S" s
 
-let tyvar ppf s =
+let tyvar_of_name s =
   if String.length s >= 2 && s.[1] = '\'' then
     (* without the space, this would be parsed as
        a character literal *)
-    Format.fprintf ppf "' %s" s
-  else Format.fprintf ppf "'%s" s
+    "' " ^ s
+  else if Keyword.is_keyword s then "'\\#" ^ s
+  else if String.equal s "_" then s
+  else "'" ^ s
 
+let tyvar ppf s = Format.fprintf ppf "%s" (tyvar_of_name s)
 let tyvar_loc f str = tyvar f str.txt
-let string_quot f x = pp f "`%s" x
+let string_quot f x = pp f "`%a" ident_of_name x
 
 (* c ['a,'b] *)
 let rec class_params_def ctxt f = function
@@ -354,8 +319,8 @@ let rec class_params_def ctxt f = function
 and type_with_label ctxt f (label, c) =
   match label with
   | Nolabel -> core_type1 ctxt f c (* otherwise parenthesize *)
-  | Labelled s -> pp f "%s:%a" s (core_type1 ctxt) c
-  | Optional s -> pp f "?%s:%a" s (core_type1 ctxt) c
+  | Labelled s -> pp f "%a:%a" ident_of_name s (core_type1 ctxt) c
+  | Optional s -> pp f "?%a:%a" ident_of_name s (core_type1 ctxt) c
 
 and core_type ctxt f x =
   if x.ptyp_attributes <> [] then
@@ -368,17 +333,14 @@ and core_type ctxt f x =
         pp f "@[<2>%a@;->@;%a@]" (* FIXME remove parens later *)
           (type_with_label ctxt) (l, ct1) (core_type ctxt) ct2
     | Ptyp_alias (ct, s) ->
-        pp f "@[<2>%a@;as@;%a@]" (core_type1 ctxt) ct tyvar s
+        pp f "@[<2>%a@;as@;%a@]" (core_type1 ctxt) ct tyvar s.txt
     | Ptyp_poly ([], ct) -> core_type ctxt f ct
     | Ptyp_poly (sl, ct) ->
         pp f "@[<2>%a%a@]"
           (fun f l ->
-            pp f "%a"
-              (fun f l ->
-                match l with
-                | [] -> ()
-                | _ -> pp f "%a@;.@;" (list tyvar_loc ~sep:"@;") l)
-              l)
+            match l with
+            | [] -> ()
+            | _ -> pp f "%a@;.@;" (list tyvar_loc ~sep:"@;") l)
           sl (core_type ctxt) ct
     | _ -> pp f "@[<2>%a@]" (core_type1 ctxt) x
 
@@ -438,8 +400,8 @@ and core_type1 ctxt f x =
           match x.pof_desc with
           | Otag (l, ct) ->
               (* Cf #7200 *)
-              pp f "@[<hov2>%s: %a@ %a@ @]" l.txt (core_type ctxt) ct
-                (attributes ctxt) x.pof_attributes
+              pp f "@[<hov2>%a: %a@ %a@ @]" ident_of_name l.txt (core_type ctxt)
+                ct (attributes ctxt) x.pof_attributes
           | Oinherit ct -> pp f "@[<hov2>%a@ @]" (core_type ctxt) ct
         in
         let field_var f = function
@@ -449,8 +411,7 @@ and core_type1 ctxt f x =
         in
         pp f "@[<hov2><@ %a%a@ > @]"
           (list core_field_type ~sep:";")
-          l field_var o
-        (* Cf #7200 *)
+          l field_var o (* Cf #7200 *)
     | Ptyp_class (li, l) ->
         (*FIXME*)
         pp f "@[<hov2>%a#%a@]"
@@ -465,8 +426,11 @@ and core_type1 ctxt f x =
         | _ ->
             pp f "@[<hov2>(module@ %a@ with@ %a)@]" longident_loc lid
               (list aux ~sep:"@ and@ ") cstrs)
+    | Ptyp_open (li, ct) ->
+        pp f "@[<hov2>%a.(%a)@]" longident_loc li (core_type ctxt) ct
     | Ptyp_extension e -> extension ctxt f e
-    | _ -> paren true (core_type ctxt) f x
+    | Ptyp_arrow _ | Ptyp_alias _ | Ptyp_poly _ ->
+        paren true (core_type ctxt) f x
 
 (********************pattern********************)
 (* be cautious when use [pattern], [pattern1] is preferred *)
@@ -478,7 +442,7 @@ and pattern ctxt f x =
   else
     match x.ppat_desc with
     | Ppat_alias (p, s) ->
-        pp f "@[<2>%a@;as@;%a@]" (pattern ctxt) p protect_ident s.txt
+        pp f "@[<2>%a@;as@;%a@]" (pattern ctxt) p ident_of_name s.txt
     | _ -> pattern_or ctxt f x
 
 and pattern_or ctxt f x =
@@ -509,19 +473,22 @@ and pattern1 ctxt (f : Format.formatter) (x : pattern) : unit =
   else
     match x.ppat_desc with
     | Ppat_variant (l, Some p) ->
-        pp f "@[<2>`%s@;%a@]" l (simple_pattern ctxt) p
-    | Ppat_construct ({ txt = Lident ("()" | "[]"); _ }, _) ->
+        pp f "@[<2>`%a@;%a@]" ident_of_name l (simple_pattern ctxt) p
+    | Ppat_construct ({ txt = Lident ("()" | "[]" | "true" | "false"); _ }, _)
+      ->
         simple_pattern ctxt f x
     | Ppat_construct (({ txt; _ } as li), po) -> (
-        if (* FIXME The third field always false *)
-           txt = Lident "::" then pp f "%a" pattern_list_helper x
+        if
+          (* FIXME The third field always false *)
+          txt = Lident "::"
+        then pp f "%a" pattern_list_helper x
         else
           match po with
           | Some ([], x) ->
               pp f "%a@;%a" longident_loc li (simple_pattern ctxt) x
           | Some (vl, x) ->
               pp f "%a@ (type %a)@;%a" longident_loc li
-                (list ~sep:"@ " string_loc)
+                (list ~sep:"@ " ident_of_name_loc)
                 vl (simple_pattern ctxt) x
           | None -> pp f "%a" longident_loc li)
     | _ -> simple_pattern ctxt f x
@@ -530,10 +497,11 @@ and simple_pattern ctxt (f : Format.formatter) (x : pattern) : unit =
   if x.ppat_attributes <> [] then pattern ctxt f x
   else
     match x.ppat_desc with
-    | Ppat_construct ({ txt = Lident (("()" | "[]") as x); _ }, None) ->
+    | Ppat_construct
+        ({ txt = Lident (("()" | "[]" | "true" | "false") as x); _ }, None) ->
         pp f "%s" x
     | Ppat_any -> pp f "_"
-    | Ppat_var { txt; _ } -> protect_ident f txt
+    | Ppat_var { txt; _ } -> ident_of_name f txt
     | Ppat_array l -> pp f "@[<2>[|%a|]@]" (list (pattern1 ctxt) ~sep:";") l
     | Ppat_unpack { txt = None } -> pp f "(module@ _)@ "
     | Ppat_unpack { txt = Some s } -> pp f "(module@ %s)@ " s
@@ -555,7 +523,7 @@ and simple_pattern ctxt (f : Format.formatter) (x : pattern) : unit =
         pp f "@[<1>(%a)@]" (list ~sep:",@;" (pattern1 ctxt)) l (* level1*)
     | Ppat_constant c -> pp f "%a" constant c
     | Ppat_interval (c1, c2) -> pp f "%a..%a" constant c1 constant c2
-    | Ppat_variant (l, None) -> pp f "`%s" l
+    | Ppat_variant (l, None) -> pp f "`%a" ident_of_name l
     | Ppat_constraint (p, ct) ->
         pp f "@[<2>(%a@;:@;%a)@]" (pattern1 ctxt) p (core_type ctxt) ct
     | Ppat_lazy p -> pp f "@[<2>(lazy@;%a)@]" (simple_pattern ctxt) p
@@ -565,7 +533,8 @@ and simple_pattern ctxt (f : Format.formatter) (x : pattern) : unit =
         let with_paren =
           match p.ppat_desc with
           | Ppat_array _ | Ppat_record _
-          | Ppat_construct ({ txt = Lident ("()" | "[]"); _ }, None) ->
+          | Ppat_construct
+              ({ txt = Lident ("()" | "[]" | "true" | "false"); _ }, None) ->
               false
           | _ -> true
         in
@@ -584,19 +553,20 @@ and label_exp ctxt f (l, opt, p) =
       | { ppat_desc = Ppat_var { txt; _ }; ppat_attributes = [] }
         when txt = rest -> (
           match opt with
-          | Some o -> pp f "?(%s=@;%a)@;" rest (expression ctxt) o
-          | None -> pp f "?%s@ " rest)
+          | Some o -> pp f "?(%a=@;%a)@;" ident_of_name rest (expression ctxt) o
+          | None -> pp f "?%a@ " ident_of_name rest)
       | _ -> (
           match opt with
           | Some o ->
-              pp f "?%s:(%a=@;%a)@;" rest (pattern1 ctxt) p (expression ctxt) o
-          | None -> pp f "?%s:%a@;" rest (simple_pattern ctxt) p))
+              pp f "?%a:(%a=@;%a)@;" ident_of_name rest (pattern1 ctxt) p
+                (expression ctxt) o
+          | None -> pp f "?%a:%a@;" ident_of_name rest (simple_pattern ctxt) p))
   | Labelled l -> (
       match p with
       | { ppat_desc = Ppat_var { txt; _ }; ppat_attributes = [] } when txt = l
         ->
-          pp f "~%s@;" l
-      | _ -> pp f "~%s:%a@;" l (simple_pattern ctxt) p)
+          pp f "~%a@;" ident_of_name l
+      | _ -> pp f "~%a:%a@;" ident_of_name l (simple_pattern ctxt) p)
 
 and sugar_expr ctxt f e =
   if e.pexp_attributes <> [] then false
@@ -649,9 +619,9 @@ and sugar_expr ctxt f e =
             | _ -> false)
         | (Lident s | Ldot (_, s)), a :: i :: rest when first_is '.' s ->
             (* extract operator:
-               assignment operators end with [right_bracket ^ "<-"],
-               access operators end with [right_bracket] directly
-            *)
+             assignment operators end with [right_bracket ^ "<-"],
+             access operators end with [right_bracket] directly
+          *)
             let multi_indices = String.contains s ';' in
             let i =
               match i.pexp_desc with
@@ -681,6 +651,35 @@ and sugar_expr ctxt f e =
         | _ -> false)
     | _ -> false
 
+and function_param ctxt f param =
+  match param.pparam_desc with
+  | Pparam_val (a, b, c) -> label_exp ctxt f (a, b, c)
+  | Pparam_newtype ty -> pp f "(type %a)@;" ident_of_name ty.txt
+
+and function_body ctxt f function_body =
+  match function_body with
+  | Pfunction_body body -> expression ctxt f body
+  | Pfunction_cases (cases, _, attrs) ->
+      pp f "@[<hv>function%a%a@]" (item_attributes ctxt) attrs (case_list ctxt)
+        cases
+
+and type_constraint ctxt f constraint_ =
+  match constraint_ with
+  | Pconstraint ty -> pp f ":@;%a" (core_type ctxt) ty
+  | Pcoerce (ty1, ty2) ->
+      pp f "%a:>@;%a"
+        (option ~first:":@;" (core_type ctxt))
+        ty1 (core_type ctxt) ty2
+
+and function_params_then_body ctxt f params constraint_ body ~delimiter =
+  pp f "%a%a%s@;%a"
+    (list (function_param ctxt) ~sep:"")
+    params
+    (option (type_constraint ctxt))
+    constraint_ delimiter
+    (function_body (under_functionrhs ctxt))
+    body
+
 and expression ctxt f x =
   if x.pexp_attributes <> [] then
     pp f "((%a)@,%a)" (expression ctxt)
@@ -688,7 +687,7 @@ and expression ctxt f x =
       (attributes ctxt) x.pexp_attributes
   else
     match x.pexp_desc with
-    | Pexp_function _ | Pexp_fun _ | Pexp_match _ | Pexp_try _ | Pexp_sequence _
+    | Pexp_function _ | Pexp_match _ | Pexp_try _ | Pexp_sequence _
     | Pexp_newtype _
       when ctxt.pipe || ctxt.semi ->
         paren true (expression reset_ctxt) f x
@@ -698,12 +697,31 @@ and expression ctxt f x =
     | Pexp_letop _
       when ctxt.semi ->
         paren true (expression reset_ctxt) f x
-    | Pexp_fun (l, e0, p, e) ->
-        pp f "@[<2>fun@;%a->@;%a@]" (label_exp ctxt) (l, e0, p)
-          (expression ctxt) e
     | Pexp_newtype (lid, e) ->
-        pp f "@[<2>fun@;(type@;%s)@;->@;%a@]" lid.txt (expression ctxt) e
-    | Pexp_function l -> pp f "@[<hv>function%a@]" (case_list ctxt) l
+        pp f "@[<2>fun@;(type@;%a)@;->@;%a@]" ident_of_name lid.txt
+          (expression ctxt) e
+    | Pexp_function (params, c, body) -> (
+        match (params, c) with
+        (* Omit [fun] if there are no params. *)
+        | [], None ->
+            (* If function cases are a direct body of a function,
+               the function node should be wrapped in parens so
+               it doesn't become part of the enclosing function. *)
+            let should_paren =
+              match body with
+              | Pfunction_cases _ -> ctxt.functionrhs
+              | Pfunction_body _ -> false
+            in
+            let ctxt' = if should_paren then reset_ctxt else ctxt in
+            pp f "@[<2>%a@]" (paren should_paren (function_body ctxt')) body
+        | [], Some c ->
+            pp f "@[<2>(%a@;%a)@]" (function_body ctxt) body
+              (type_constraint ctxt) c
+        | _ :: _, _ ->
+            pp f "@[<2>fun@;%a@]"
+              (fun f () ->
+                function_params_then_body ctxt f params c body ~delimiter:"->")
+              ())
     | Pexp_match (e, l) ->
         pp f "@[<hv0>@[<hv0>@[<2>match %a@]@ with@]%a@]" (expression reset_ctxt)
           e (case_list ctxt) l
@@ -741,7 +759,7 @@ and expression ctxt f x =
                   &&
                   match l with
                   (* See #7200: avoid turning (~- 1) into (- 1) which is
-                     parsed as an int literal *)
+                       parsed as an int literal *)
                   | [ (_, { pexp_desc = Pexp_constant _ }) ] -> false
                   | _ -> true
                 then String.sub s 1 (String.length s - 1)
@@ -758,9 +776,9 @@ and expression ctxt f x =
                 (fun f (e, l) ->
                   pp f "%a@ %a" (expression2 ctxt) e
                     (list (label_x_expression_param reset_ctxt))
-                    l
+                    l)
                   (* reset here only because [function,match,try,sequence]
-                     are lower priority *))
+                       are lower priority *)
                 (e, l))
     | Pexp_construct (li, Some eo) when not (is_simple_construct (view_expr x))
       -> (
@@ -783,8 +801,7 @@ and expression ctxt f x =
             match eo with
             | Some x ->
                 pp f "@;@[<2>else@;%a@]" (expression (under_semi ctxt)) x
-            | None -> ()
-            (* pp f "()" *))
+            | None -> () (* pp f "()" *))
           eo
     | Pexp_sequence _ ->
         let rec sequence_helper acc = function
@@ -796,11 +813,11 @@ and expression ctxt f x =
         pp f "@[<hv>%a@]" (list (expression (under_semi ctxt)) ~sep:";@;") lst
     | Pexp_new li -> pp f "@[<hov2>new@ %a@]" longident_loc li
     | Pexp_setinstvar (s, e) ->
-        pp f "@[<hov2>%s@ <-@ %a@]" s.txt (expression ctxt) e
+        pp f "@[<hov2>%a@ <-@ %a@]" ident_of_name s.txt (expression ctxt) e
     | Pexp_override l ->
         (* FIXME *)
         let string_x_expression f (s, e) =
-          pp f "@[<hov2>%s@ =@ %a@]" s.txt (expression ctxt) e
+          pp f "@[<hov2>%a@ =@ %a@]" ident_of_name s.txt (expression ctxt) e
         in
         pp f "@[<hov2>{<%a>}@]" (list string_x_expression ~sep:";") l
     | Pexp_letmodule (s, me, e) ->
@@ -823,7 +840,8 @@ and expression ctxt f x =
         pp f "@[<2>let open%s %a in@;%a@]"
           (override o.popen_override)
           (module_expr ctxt) o.popen_expr (expression ctxt) e
-    | Pexp_variant (l, Some eo) -> pp f "@[<2>`%s@;%a@]" l (simple_expr ctxt) eo
+    | Pexp_variant (l, Some eo) ->
+        pp f "@[<2>`%a@;%a@]" ident_of_name l (simple_expr ctxt) eo
     | Pexp_letop { let_; ands; body } ->
         pp f "@[<2>@[<v>%a@,%a@] in@;<1 -2>%a@]" (binding_op ctxt) let_
           (list ~sep:"@," (binding_op ctxt))
@@ -846,7 +864,8 @@ and expression2 ctxt f x =
     match x.pexp_desc with
     | Pexp_field (e, li) ->
         pp f "@[<hov2>%a.%a@]" (simple_expr ctxt) e longident_loc li
-    | Pexp_send (e, s) -> pp f "@[<hov2>%a#%s@]" (simple_expr ctxt) e s.txt
+    | Pexp_send (e, s) ->
+        pp f "@[<hov2>%a#%a@]" (simple_expr ctxt) e ident_of_name s.txt
     | _ -> simple_expr ctxt f x
 
 and simple_expr ctxt f x =
@@ -857,6 +876,8 @@ and simple_expr ctxt f x =
         match view_expr x with
         | `nil -> pp f "[]"
         | `tuple -> pp f "()"
+        | `btrue -> pp f "true"
+        | `bfalse -> pp f "false"
         | `list xs ->
             pp f "@[<hv0>[%a]@]"
               (list (expression (under_semi ctxt)) ~sep:";@;")
@@ -876,10 +897,9 @@ and simple_expr ctxt f x =
     | Pexp_coerce (e, cto1, ct) ->
         pp f "(%a%a :> %a)" (expression ctxt) e
           (option (core_type ctxt) ~first:" : " ~last:" ")
-          cto1
-          (* no sep hint*) (core_type ctxt)
-          ct
-    | Pexp_variant (l, None) -> pp f "`%s" l
+          cto1 (* no sep hint*)
+          (core_type ctxt) ct
+    | Pexp_variant (l, None) -> pp f "`%a" ident_of_name l
     | Pexp_record (l, eo) ->
         let longident_x_expression f (li, e) =
           match e with
@@ -947,11 +967,13 @@ and class_type_field ctxt f x =
       pp f "@[<2>inherit@ %a@]%a" (class_type ctxt) ct (item_attributes ctxt)
         x.pctf_attributes
   | Pctf_val (s, mf, vf, ct) ->
-      pp f "@[<2>val @ %a%a%s@ :@ %a@]%a" mutable_flag mf virtual_flag vf s.txt
-        (core_type ctxt) ct (item_attributes ctxt) x.pctf_attributes
+      pp f "@[<2>val @ %a%a%a@ :@ %a@]%a" mutable_flag mf virtual_flag vf
+        ident_of_name s.txt (core_type ctxt) ct (item_attributes ctxt)
+        x.pctf_attributes
   | Pctf_method (s, pf, vf, ct) ->
-      pp f "@[<2>method %a %a%s :@;%a@]%a" private_flag pf virtual_flag vf s.txt
-        (core_type ctxt) ct (item_attributes ctxt) x.pctf_attributes
+      pp f "@[<2>method %a %a%a :@;%a@]%a" private_flag pf virtual_flag vf
+        ident_of_name s.txt (core_type ctxt) ct (item_attributes ctxt)
+        x.pctf_attributes
   | Pctf_constraint (ct1, ct2) ->
       pp f "@[<2>constraint@ %a@ =@ %a@]%a" (core_type ctxt) ct1
         (core_type ctxt) ct2 (item_attributes ctxt) x.pctf_attributes
@@ -997,8 +1019,8 @@ and class_type ctxt f x =
 and class_type_declaration_list ctxt f l =
   let class_type_declaration kwd f x =
     let { pci_params = ls; pci_name = { txt; _ }; _ } = x in
-    pp f "@[<2>%s %a%a%s@ =@ %a@]%a" kwd virtual_flag x.pci_virt
-      (class_params_def ctxt) ls txt (class_type ctxt) x.pci_expr
+    pp f "@[<2>%s %a%a%a@ =@ %a@]%a" kwd virtual_flag x.pci_virt
+      (class_params_def ctxt) ls ident_of_name txt (class_type ctxt) x.pci_expr
       (item_attributes ctxt) x.pci_attributes
   in
   match l with
@@ -1016,16 +1038,19 @@ and class_field ctxt f x =
   | Pcf_inherit (ovf, ce, so) ->
       pp f "@[<2>inherit@ %s@ %a%a@]%a" (override ovf) (class_expr ctxt) ce
         (fun f so ->
-          match so with None -> () | Some s -> pp f "@ as %s" s.txt)
+          match so with
+          | None -> ()
+          | Some s -> pp f "@ as %a" ident_of_name s.txt)
         so (item_attributes ctxt) x.pcf_attributes
   | Pcf_val (s, mf, Cfk_concrete (ovf, e)) ->
-      pp f "@[<2>val%s %a%s =@;%a@]%a" (override ovf) mutable_flag mf s.txt
-        (expression ctxt) e (item_attributes ctxt) x.pcf_attributes
+      pp f "@[<2>val%s %a%a =@;%a@]%a" (override ovf) mutable_flag mf
+        ident_of_name s.txt (expression ctxt) e (item_attributes ctxt)
+        x.pcf_attributes
   | Pcf_method (s, pf, Cfk_virtual ct) ->
-      pp f "@[<2>method virtual %a %s :@;%a@]%a" private_flag pf s.txt
-        (core_type ctxt) ct (item_attributes ctxt) x.pcf_attributes
+      pp f "@[<2>method virtual %a %a :@;%a@]%a" private_flag pf ident_of_name
+        s.txt (core_type ctxt) ct (item_attributes ctxt) x.pcf_attributes
   | Pcf_val (s, mf, Cfk_virtual ct) ->
-      pp f "@[<2>val virtual %a%s :@ %a@]%a" mutable_flag mf s.txt
+      pp f "@[<2>val virtual %a%a :@ %a@]%a" mutable_flag mf ident_of_name s.txt
         (core_type ctxt) ct (item_attributes ctxt) x.pcf_attributes
   | Pcf_method (s, pf, Cfk_concrete (ovf, e)) ->
       let bind e =
@@ -1039,6 +1064,7 @@ and class_field ctxt f x =
                 ppat_attributes = [];
               };
             pvb_expr = e;
+            pvb_constraint = None;
             pvb_attributes = [];
             pvb_loc = Location.none;
           }
@@ -1046,7 +1072,8 @@ and class_field ctxt f x =
       pp f "@[<2>method%s %a%a@]%a" (override ovf) private_flag pf
         (fun f -> function
           | { pexp_desc = Pexp_poly (e, Some ct); pexp_attributes = []; _ } ->
-              pp f "%s :@;%a=@;%a" s.txt (core_type ctxt) ct (expression ctxt) e
+              pp f "%a :@;%a=@;%a" ident_of_name s.txt (core_type ctxt) ct
+                (expression ctxt) e
           | { pexp_desc = Pexp_poly (e, None); pexp_attributes = []; _ } ->
               bind e
           | _ -> bind e)
@@ -1112,7 +1139,7 @@ and module_type ctxt f x =
   else
     match x.pmty_desc with
     | Pmty_functor (Unit, mt2) ->
-        pp f "@[<hov2>functor () ->@ %a@]" (module_type ctxt) mt2
+        pp f "@[<hov2>() ->@ %a@]" (module_type ctxt) mt2
     | Pmty_functor (Named (s, mt1), mt2) -> (
         match s.txt with
         | None ->
@@ -1130,19 +1157,15 @@ and module_type ctxt f x =
 
 and with_constraint ctxt f = function
   | Pwith_type (li, ({ ptype_params = ls; _ } as td)) ->
-      let ls = List.map fst ls in
-      pp f "type@ %a %a =@ %a"
-        (list (core_type ctxt) ~sep:"," ~first:"(" ~last:")")
-        ls longident_loc li (type_declaration ctxt) td
+      pp f "type@ %a %a =@ %a" (type_params ctxt) ls longident_loc li
+        (type_declaration ctxt) td
   | Pwith_module (li, li2) ->
       pp f "module %a =@ %a" longident_loc li longident_loc li2
   | Pwith_modtype (li, mty) ->
       pp f "module type %a =@ %a" longident_loc li (module_type ctxt) mty
   | Pwith_typesubst (li, ({ ptype_params = ls; _ } as td)) ->
-      let ls = List.map fst ls in
-      pp f "type@ %a %a :=@ %a"
-        (list (core_type ctxt) ~sep:"," ~first:"(" ~last:")")
-        ls longident_loc li (type_declaration ctxt) td
+      pp f "type@ %a %a :=@ %a" (type_params ctxt) ls longident_loc li
+        (type_declaration ctxt) td
   | Pwith_modsubst (li, li2) ->
       pp f "module %a :=@ %a" longident_loc li longident_loc li2
   | Pwith_modtypesubst (li, mty) ->
@@ -1157,8 +1180,7 @@ and module_type1 ctxt f x =
     | Pmty_signature s ->
         pp f "@[<hv0>@[<hv2>sig@ %a@]@ end@]" (* "@[<hov>sig@ %a@ end@]" *)
           (list (signature_item ctxt))
-          s
-        (* FIXME wrong indentation*)
+          s (* FIXME wrong indentation*)
     | Pmty_typeof me ->
         pp f "@[<hov2>module@ type@ of@ %a@]" (module_expr ctxt) me
     | Pmty_extension e -> extension ctxt f e
@@ -1176,16 +1198,16 @@ and signature_item ctxt f x : unit =
       type_def_list ctxt f (Recursive, false, l)
   | Psig_value vd ->
       let intro = if vd.pval_prim = [] then "val" else "external" in
-      pp f "@[<2>%s@ %a@ :@ %a@]%a" intro protect_ident vd.pval_name.txt
+      pp f "@[<2>%s@ %a@ :@ %a@]%a" intro ident_of_name vd.pval_name.txt
         (value_description ctxt) vd (item_attributes ctxt) vd.pval_attributes
   | Psig_typext te -> type_extension ctxt f te
   | Psig_exception ed -> exception_declaration ctxt f ed
   | Psig_class l -> (
       let class_description kwd f
           ({ pci_params = ls; pci_name = { txt; _ }; _ } as x) =
-        pp f "@[<2>%s %a%a%s@;:@;%a@]%a" kwd virtual_flag x.pci_virt
-          (class_params_def ctxt) ls txt (class_type ctxt) x.pci_expr
-          (item_attributes ctxt) x.pci_attributes
+        pp f "@[<2>%s %a%a%a@;:@;%a@]%a" kwd virtual_flag x.pci_virt
+          (class_params_def ctxt) ls ident_of_name txt (class_type ctxt)
+          x.pci_expr (item_attributes ctxt) x.pci_attributes
       in
       match l with
       | [] -> ()
@@ -1281,6 +1303,7 @@ and module_expr ctxt f x =
     | Pmod_apply (me1, me2) ->
         pp f "(%a)(%a)" (module_expr ctxt) me1 (module_expr ctxt) me2
         (* Cf: #7200 *)
+    | Pmod_apply_unit me1 -> pp f "(%a)()" (module_expr ctxt) me1
     | Pmod_unpack e -> pp f "(val@ %a)" (expression ctxt) e
     | Pmod_extension e -> extension ctxt f e
 
@@ -1306,88 +1329,43 @@ and payload ctxt f = function
       expression ctxt f e
 
 (* transform [f = fun g h -> ..] to [f g h = ... ] could be improved *)
-and binding ctxt f { pvb_pat = p; pvb_expr = x; _ } =
+and binding ctxt f { pvb_pat = p; pvb_expr = x; pvb_constraint = ct; _ } =
   (* .pvb_attributes have already been printed by the caller, #bindings *)
   let rec pp_print_pexp_function f x =
     if x.pexp_attributes <> [] then pp f "=@;%a" (expression ctxt) x
     else
       match x.pexp_desc with
-      | Pexp_fun (label, eo, p, e) ->
-          if label = Nolabel then
-            pp f "%a@ %a" (simple_pattern ctxt) p pp_print_pexp_function e
-          else
-            pp f "%a@ %a" (label_exp ctxt) (label, eo, p) pp_print_pexp_function
-              e
+      | Pexp_function (params, c, body) ->
+          function_params_then_body ctxt f params c body ~delimiter:"="
       | Pexp_newtype (str, e) ->
-          pp f "(type@ %s)@ %a" str.txt pp_print_pexp_function e
+          pp f "(type@ %a)@ %a" ident_of_name str.txt pp_print_pexp_function e
       | _ -> pp f "=@;%a" (expression ctxt) x
   in
-  let tyvars_str tyvars = List.map (fun v -> v.txt) tyvars in
-  let is_desugared_gadt p e =
-    let gadt_pattern =
-      match p with
-      | {
-       ppat_desc =
-         Ppat_constraint
-           ( ({ ppat_desc = Ppat_var _ } as pat),
-             { ptyp_desc = Ptyp_poly (args_tyvars, rt) } );
-       ppat_attributes = [];
-      } ->
-          Some (pat, args_tyvars, rt)
-      | _ -> None
-    in
-    let rec gadt_exp tyvars e =
-      match e with
-      | { pexp_desc = Pexp_newtype (tyvar, e); pexp_attributes = [] } ->
-          gadt_exp (tyvar :: tyvars) e
-      | { pexp_desc = Pexp_constraint (e, ct); pexp_attributes = [] } ->
-          Some (List.rev tyvars, e, ct)
-      | _ -> None
-    in
-    let gadt_exp = gadt_exp [] e in
-    match (gadt_pattern, gadt_exp) with
-    | Some (p, pt_tyvars, pt_ct), Some (e_tyvars, e, e_ct)
-      when tyvars_str pt_tyvars = tyvars_str e_tyvars ->
-        let ety = varify_type_constructors e_tyvars e_ct in
-        if ety = pt_ct then Some (p, pt_tyvars, e_ct, e) else None
-    | _ -> None
-  in
-  if x.pexp_attributes <> [] then
-    match p with
-    | {
-     ppat_desc =
-       Ppat_constraint
-         ( ({ ppat_desc = Ppat_var _; _ } as pat),
-           ({ ptyp_desc = Ptyp_poly _; _ } as typ) );
-     ppat_attributes = [];
-     _;
-    } ->
-        pp f "%a@;: %a@;=@;%a" (simple_pattern ctxt) pat (core_type ctxt) typ
-          (expression ctxt) x
-    | _ -> pp f "%a@;=@;%a" (pattern ctxt) p (expression ctxt) x
-  else
-    match is_desugared_gadt p x with
-    | Some (p, [], ct, e) ->
-        pp f "%a@;: %a@;=@;%a" (simple_pattern ctxt) p (core_type ctxt) ct
-          (expression ctxt) e
-    | Some (p, tyvars, ct, e) ->
-        pp f "%a@;: type@;%a.@;%a@;=@;%a" (simple_pattern ctxt) p
-          (list pp_print_string ~sep:"@;")
-          (tyvars_str tyvars) (core_type ctxt) ct (expression ctxt) e
-    | None -> (
-        match p with
-        | { ppat_desc = Ppat_constraint (p, ty); ppat_attributes = [] } -> (
-            (* special case for the first*)
-            match ty with
-            | { ptyp_desc = Ptyp_poly _; ptyp_attributes = [] } ->
-                pp f "%a@;:@;%a@;=@;%a" (simple_pattern ctxt) p (core_type ctxt)
-                  ty (expression ctxt) x
-            | _ ->
-                pp f "(%a@;:@;%a)@;=@;%a" (simple_pattern ctxt) p
-                  (core_type ctxt) ty (expression ctxt) x)
-        | { ppat_desc = Ppat_var _; ppat_attributes = [] } ->
-            pp f "%a@ %a" (simple_pattern ctxt) p pp_print_pexp_function x
-        | _ -> pp f "%a@;=@;%a" (pattern ctxt) p (expression ctxt) x)
+  match (ct, p) with
+  | ( None,
+      {
+        ppat_attributes = [];
+        ppat_desc =
+          Ppat_constraint
+            (({ ppat_desc = Ppat_var _; ppat_attributes = [] } as p), typ);
+      } )
+  | Some (Pvc_constraint { locally_abstract_univars = []; typ }), p ->
+      pp f "%a@;:@;%a@;=@;%a" (simple_pattern ctxt) p (core_type ctxt) typ
+        (expression ctxt) x
+  | Some (Pvc_constraint { locally_abstract_univars = vars; typ }), _ ->
+      pp f "%a@;: type@;%a.@;%a@;=@;%a" (simple_pattern ctxt) p
+        (list pp_print_string ~sep:"@;")
+        (List.map (fun x -> x.txt) vars)
+        (core_type ctxt) typ (expression ctxt) x
+  | Some (Pvc_coercion { ground = None; coercion }), _ ->
+      pp f "%a@;:>@;%a@;=@;%a" (simple_pattern ctxt) p (core_type ctxt) coercion
+        (expression ctxt) x
+  | Some (Pvc_coercion { ground = Some ground; coercion }), _ ->
+      pp f "%a@;:%a@;:>@;%a@;=@;%a" (simple_pattern ctxt) p (core_type ctxt)
+        ground (core_type ctxt) coercion (expression ctxt) x
+  | None, { ppat_desc = Ppat_var _; ppat_attributes = [] } ->
+      pp f "%a@ %a" (simple_pattern ctxt) p pp_print_pexp_function x
+  | _, _ -> pp f "%a@;=@;%a" (pattern ctxt) p (expression ctxt) x
 
 (* [in] is not printed *)
 and bindings ctxt f (rf, l) =
@@ -1489,8 +1467,8 @@ and structure_item ctxt f x =
       let class_declaration kwd f
           ({ pci_params = ls; pci_name = { txt; _ }; _ } as x) =
         let args, constr, cl = extract_class_args x.pci_expr in
-        pp f "@[<2>%s %a%a%s %a%a=@;%a@]%a" kwd virtual_flag x.pci_virt
-          (class_params_def ctxt) ls txt
+        pp f "@[<2>%s %a%a%a %a%a=@;%a@]%a" kwd virtual_flag x.pci_virt
+          (class_params_def ctxt) ls ident_of_name txt
           (list (label_exp ctxt))
           args (option class_constraint) constr (class_expr ctxt) cl
           (item_attributes ctxt) x.pci_attributes
@@ -1506,7 +1484,7 @@ and structure_item ctxt f x =
             xs)
   | Pstr_class_type l -> class_type_declaration_list ctxt f l
   | Pstr_primitive vd ->
-      pp f "@[<hov2>external@ %a@ :@ %a@]%a" protect_ident vd.pval_name.txt
+      pp f "@[<hov2>external@ %a@ :@ %a@]%a" ident_of_name vd.pval_name.txt
         (value_description ctxt) vd (item_attributes ctxt) vd.pval_attributes
   | Pstr_include incl ->
       pp f "@[<hov2>include@ %a@]%a" (module_expr ctxt) incl.pincl_mod
@@ -1561,8 +1539,8 @@ and type_def_list ctxt f (rf, exported, l) =
       else if exported then " ="
       else " :="
     in
-    pp f "@[<2>%s %a%a%s%s%a@]%a" kwd nonrec_flag rf (type_params ctxt)
-      x.ptype_params x.ptype_name.txt eq (type_declaration ctxt) x
+    pp f "@[<2>%s %a%a%a%s%a@]%a" kwd nonrec_flag rf (type_params ctxt)
+      x.ptype_params ident_of_name x.ptype_name.txt eq (type_declaration ctxt) x
       (item_attributes ctxt) x.ptype_attributes
   in
   match l with
@@ -1575,8 +1553,9 @@ and type_def_list ctxt f (rf, exported, l) =
 
 and record_declaration ctxt f lbls =
   let type_record_field f pld =
-    pp f "@[<2>%a%s:@;%a@;%a@]" mutable_flag pld.pld_mutable pld.pld_name.txt
-      (core_type ctxt) pld.pld_type (attributes ctxt) pld.pld_attributes
+    pp f "@[<2>%a%a:@;%a@;%a@]" mutable_flag pld.pld_mutable ident_of_name
+      pld.pld_name.txt (core_type ctxt) pld.pld_type (attributes ctxt)
+      pld.pld_attributes
   in
   pp f "{@\n%a}" (list type_record_field ~sep:";@\n") lbls
 
@@ -1697,11 +1676,11 @@ and label_x_expression_param ctxt f (l, e) =
   match l with
   | Nolabel -> expression2 ctxt f e (* level 2*)
   | Optional str ->
-      if Some str = simple_name then pp f "?%s" str
-      else pp f "?%s:%a" str (simple_expr ctxt) e
+      if Some str = simple_name then pp f "?%a" ident_of_name str
+      else pp f "?%a:%a" ident_of_name str (simple_expr ctxt) e
   | Labelled lbl ->
-      if Some lbl = simple_name then pp f "~%s" lbl
-      else pp f "~%s:%a" lbl (simple_expr ctxt) e
+      if Some lbl = simple_name then pp f "~%a" ident_of_name lbl
+      else pp f "~%a:%a" ident_of_name lbl (simple_expr ctxt) e
 
 and directive_argument f x =
   match x.pdira_desc with
@@ -1746,13 +1725,15 @@ let core_type = core_type reset_ctxt
 let pattern = pattern reset_ctxt
 let signature = signature reset_ctxt
 let structure = structure reset_ctxt
-let class_expr = class_expr reset_ctxt
-let class_field = class_field reset_ctxt
-let class_type = class_type reset_ctxt
-let class_signature = class_signature reset_ctxt
-let class_type_field = class_type_field reset_ctxt
 let module_expr = module_expr reset_ctxt
 let module_type = module_type reset_ctxt
-let signature_item = signature_item reset_ctxt
+let class_field = class_field reset_ctxt
+let class_type_field = class_type_field reset_ctxt
+let class_expr = class_expr reset_ctxt
+let class_type = class_type reset_ctxt
 let structure_item = structure_item reset_ctxt
+let signature_item = signature_item reset_ctxt
+let binding = binding reset_ctxt
+let payload = payload reset_ctxt
+let class_signature = class_signature reset_ctxt
 let type_declaration = type_declaration reset_ctxt
