@@ -48,6 +48,7 @@ module Generator : sig
   val make :
     ?attributes:Attribute.packed list (* deprecated, unused *) ->
     ?deps:deriver list ->
+    ?unused_code_warnings:bool ->
     ('f, 'output_ast) Args.t ->
     (loc:Location.t -> path:string -> 'input_ast -> 'f) ->
     ('output_ast, 'input_ast) t
@@ -57,11 +58,16 @@ module Generator : sig
       [deps] is a list of derivers that this generator depends on.
 
       [attributes] is deprecated and unused. It is only kept for backward
-      compatibility. *)
+      compatibility.
+
+      [unused_code_warning] controls whether unused code warnings (e.g. warnings
+      32 and 60) are enabled or suppressed in emitted code. Default is [false],
+      which suppresses the warnings. *)
 
   val make_noarg :
     ?attributes:Attribute.packed list (* deprecated, unused *) ->
     ?deps:deriver list ->
+    ?unused_code_warnings:bool ->
     (loc:Location.t -> path:string -> 'input_ast -> 'output_ast) ->
     ('output_ast, 'input_ast) t
   (** Same as {!make}, but without arguments. *)
@@ -70,6 +76,7 @@ module Generator : sig
     val make :
       ?attributes:Attribute.packed list (* deprecated, unused *) ->
       ?deps:deriver list ->
+      ?unused_code_warnings:bool ->
       ('f, 'output_ast) Args.t ->
       (ctxt:Expansion_context.Deriver.t -> 'input_ast -> 'f) ->
       ('output_ast, 'input_ast) t
@@ -79,6 +86,7 @@ module Generator : sig
     val make_noarg :
       ?attributes:Attribute.packed list (* deprecated, unused *) ->
       ?deps:deriver list ->
+      ?unused_code_warnings:bool ->
       (ctxt:Expansion_context.Deriver.t -> 'input_ast -> 'output_ast) ->
       ('output_ast, 'input_ast) t
     (** Same as {!Generator.make_noarg}, but the generator has access to an
@@ -97,10 +105,12 @@ with type deriver := t
 
 val add :
   ?str_type_decl:(structure, rec_flag * type_declaration list) Generator.t ->
+  ?str_class_type_decl:(structure, class_type_declaration list) Generator.t ->
   ?str_type_ext:(structure, type_extension) Generator.t ->
   ?str_exception:(structure, type_exception) Generator.t ->
   ?str_module_type_decl:(structure, module_type_declaration) Generator.t ->
   ?sig_type_decl:(signature, rec_flag * type_declaration list) Generator.t ->
+  ?sig_class_type_decl:(signature, class_type_declaration list) Generator.t ->
   ?sig_type_ext:(signature, type_extension) Generator.t ->
   ?sig_exception:(signature, type_exception) Generator.t ->
   ?sig_module_type_decl:(signature, module_type_declaration) Generator.t ->
@@ -123,10 +133,12 @@ val add :
 val add_alias :
   string ->
   ?str_type_decl:t list ->
+  ?str_class_type_decl:t list ->
   ?str_type_ext:t list ->
   ?str_exception:t list ->
   ?str_module_type_decl:t list ->
   ?sig_type_decl:t list ->
+  ?sig_class_type_decl:t list ->
   ?sig_type_ext:t list ->
   ?sig_exception:t list ->
   ?sig_module_type_decl:t list ->
@@ -139,5 +151,4 @@ val add_alias :
 
 val ignore : t -> unit
 (** Ignore a deriver. So that one can write:
-    [Deriving.add ... |>
-    Deriving.ignore] *)
+    [Deriving.add ... |> Deriving.ignore] *)
