@@ -24,6 +24,8 @@ let pretty = ref false
 let styler = ref None
 let output_metadata_filename = ref None
 let corrected_suffix = ref ".ppx-corrected"
+let keywords = ref None
+let raise_embedded_errors_flag = ref false
 
 let ghost =
   object
@@ -31,6 +33,17 @@ let ghost =
     method! location loc = { loc with loc_ghost = true }
   end
 
+let raise_embedded_errors =
+  object
+    inherit Ast_traverse.map as super
+
+    method! extension extension =
+      if !raise_embedded_errors_flag then
+        extension |> Location.Error.of_extension
+        |> Option.iter ~f:Location.Error.raise;
+      super#extension extension
+  end
+
 let chop_prefix ~prefix x =
   if String.is_prefix ~prefix x then
     Some (String.drop_prefix x (String.length prefix))
@@ -695,7 +708,9 @@ let map_structure_gen st ~tool_name ~hook ~expect_mismatch_handler ~input_name
       ~dropped_so_far:Attribute.dropped_so_far_structure ~hook
       ~expect_mismatch_handler ~input_name ~embed_errors
   in
-  st |> lint lint_errors |> cookies_and_check |> with_errors (List.rev errors)
+  st |> lint lint_errors |> cookies_and_check
+  |> with_errors (List.rev errors)
+  |> raise_embedded_errors#structure
 
 let map_structure st =
   match
@@ -771,7 +786,9 @@ let map_signature_gen sg ~tool_name ~hook ~expect_mismatch_handler ~input_name
       ~dropped_so_far:Attribute.dropped_so_far_signature ~hook
       ~expect_mismatch_handler ~input_name ~embed_errors
   in
-  sg |> lint lint_errors |> cookies_and_check |> with_errors (List.rev errors)
+  sg |> lint lint_errors |> cookies_and_check
+  |> with_errors (List.rev errors)
+  |> raise_embedded_errors#signature
 
 let map_signature sg =
   match
@@ -1058,8 +1075,8 @@ let process_ast (ast : Intf_or_impl.t) ~input_name ~tool_name ~hook
       in
       Intf_or_impl.Impl ast
 
-let process_file (kind : Kind.t) fn ~input_name ~relocate ~output_mode
-    ~embed_errors ~output =
+let process_file (kind : Kind.t) fn ~input_name ~relocate ~use_compiler_pprint
+    ~output_mode ~embed_errors ~output =
   File_property.reset_all ();
   List.iter (List.rev !process_file_hooks) ~f:(fun f -> f ());
   corrections := [];
@@ -1128,7 +1145,7 @@ let process_file (kind : Kind.t) fn ~input_name ~relocate ~output_mode
         Reconcile.reconcile corrections
           ~contents:(Lazy.force input_contents)
           ~output:(Some corrected) ~input_filename:fn ~input_name
-          ~target:Corrected ?styler:!styler ~kind;
+          ~target:Corrected ?styler:!styler ~kind ~use_compiler_pprint;
         true
   in
 
@@ -1137,9 +1154,11 @@ let process_file (kind : Kind.t) fn ~input_name ~relocate ~output_mode
   | Pretty_print ->
       with_output output ~binary:false ~f:(fun oc ->
           let ppf = Stdlib.Format.formatter_of_out_channel oc in
-          (match ast with
-          | Intf ast -> Pprintast.signature ppf ast
-          | Impl ast -> Pprintast.structure ppf ast);
+          (if use_compiler_pprint then Utils.print_as_compiler_source ppf ast
+           else
+             match ast with
+             | Intf ast -> Pprintast.signature ppf ast
+             | Impl ast -> Pprintast.structure ppf ast);
           let null_ast =
             match ast with Intf [] | Impl [] -> true | _ -> false
           in
@@ -1154,14 +1173,14 @@ let process_file (kind : Kind.t) fn ~input_name ~relocate ~output_mode
           let ppf = Stdlib.Format.formatter_of_out_channel oc in
           let ast = add_cookies ast in
           (match ast with
-          | Intf ast -> Sexp.pp_hum ppf (Ast_traverse.sexp_of#signature ast)
-          | Impl ast -> Sexp.pp_hum ppf (Ast_traverse.sexp_of#structure ast));
+          | Intf ast -> Pp_ast.signature ppf ast
+          | Impl ast -> Pp_ast.structure ppf ast);
           Stdlib.Format.pp_print_newline ppf ())
   | Reconcile mode ->
       Reconcile.reconcile !replacements
         ~contents:(Lazy.force input_contents)
         ~output ~input_filename:fn ~input_name ~target:(Output mode)
-        ?styler:!styler ~kind);
+        ?styler:!styler ~kind ~use_compiler_pprint);
 
   if
     mismatches_found && match !diff_command with Some "-" -> false | _ -> true
@@ -1175,6 +1194,7 @@ let output = ref None
 let kind = ref None
 let input = ref None
 let embed_errors = ref false
+let use_compiler_pprint = ref false
 
 let set_input fn =
   match !input with
@@ -1316,6 +1336,9 @@ let shared_args =
        applied before all impl and intf." );
     ("-cookie", Arg.String set_cookie, "NAME=EXPR Set the cookie NAME to EXPR");
     ("--cookie", Arg.String set_cookie, " Same as -cookie");
+    ( "-raise-embedded-errors",
+      Arg.Set raise_embedded_errors_flag,
+      " Raise the first embedded error found in the processed AST" );
   ]
 
 let () =
@@ -1409,6 +1432,18 @@ let standalone_args =
     ( "-corrected-suffix",
       Arg.Set_string corrected_suffix,
       "SUFFIX Suffix to append to corrected files" );
+    ( "-keywords",
+      Arg.String (fun s -> keywords := Some s),
+      "<version+list> Set keywords according to the version+list \
+       specification. Allows using a set of keywords different from the one of \
+       the current compiler for backward compatibility." );
+    ( "--keywords",
+      Arg.String (fun s -> keywords := Some s),
+      "<version+list> Same as -keywords" );
+    ( "--use-compiler-pp",
+      Arg.Set use_compiler_pprint,
+      "Force migrating the AST back to the compiler's version before printing \
+       it as source code using the compiler's Pprintast utilities." );
   ]
 
 let get_args ?(standalone_args = standalone_args) () =
@@ -1418,6 +1453,7 @@ let standalone_main () =
   let usage = Printf.sprintf "%s [extra_args] [<files>]" exe_name in
   let args = get_args () in
   Arg.parse (Arg.align args) set_input usage;
+  Astlib.Keyword.apply_keyword_edition ~cli:!keywords ();
   interpret_mask ();
   if !request_print_transformations then (
     print_transformations ();
@@ -1447,6 +1483,7 @@ let standalone_main () =
       in
       process_file kind fn ~input_name ~relocate ~output_mode:!output_mode
         ~output:!output ~embed_errors:!embed_errors
+        ~use_compiler_pprint:!use_compiler_pprint
 
 let rewrite_binary_ast_file input_fn output_fn =
   let input_name, input_version, ast = load_input_run_as_ppx input_fn in
@@ -1544,3 +1581,19 @@ let enable_checks () =
 let enable_location_check () = perform_locations_check := true
 let disable_location_check () = perform_locations_check := false
 let map_structure st = map_structure st
+
+let () =
+  register_transformation "expand_inline"
+    ~rules:
+      [
+        Context_free.Rule.attr_str_floating_expect_and_expand
+          (Attribute.Floating.declare "expand_inline" Structure_item
+             Ast_pattern.(pstr __)
+             Fn.id)
+          (fun ~ctxt:_ items -> Utils.prettify_odoc_attributes#structure items);
+        Context_free.Rule.attr_sig_floating_expect_and_expand
+          (Attribute.Floating.declare "expand_inline" Signature_item
+             Ast_pattern.(psig __)
+             Fn.id)
+          (fun ~ctxt:_ items -> Utils.prettify_odoc_attributes#signature items);
+      ]
